import functools
import pathlib

import jinja2
import numpy as np

from . import utils


class BaseKernelRunner:
    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
    ):
        self.pixels_x = pixels_x
        self.pixels_y = pixels_y
        self.frames = frames
        if constant_memory_cells == 0:
            # if not set, guess same as input; may save one recompilation
            self.constant_memory_cells = frames
        else:
            self.constant_memory_cells = constant_memory_cells
        # preview will only be single memory cell
        self.preview_shape = (self.pixels_x, self.pixels_y)
        self.input_data_dtype = input_data_dtype
        self.output_data_dtype = output_data_dtype

    def correct(self, flags):
        """Correct (already loaded) image data according to flags

        Detector-specific subclass must define this method. It should assume that image
        data, cell table, and other data (including constants) has already been loaded.
        It should probably run some {CPU,GPU} kernel and output should go into
        self.processed_data{,_gpu}.

        Keep in mind that user only gets output from compute_preview or reshape
        (either of these should come after correct).

        The submodules providing subclasses should have some IntFlag enums defining
        which flags are available to pass along to the kernel. A zero flag should allow
        the kernel to do no actual correction - but still copy the data between buffers
        and cast it to desired output type.

        """
        raise NotImplementedError()

    @property
    def preview_data_views(self):
        """Should provide tuple of views of data for different preview layers - as a
        minimum, raw and corrected.  View is expected to have axis order frame, slow
        scan, fast scan with the latter two matching preview_shape (which is X and which
        is Y may differ between detectors)."""

        raise NotImplementedError()

    def flush_buffers(self, constants=None):
        """Optional reset internal buffers (implement in subclasses which need this)"""
        pass

    def compute_previews(self, preview_index):
        """Generate single slice or reduction previews for raw and corrected data and
        any other layers, determined by self.preview_data_views

        Special values of preview_index are -1 for max, -2 for mean, -3 for sum, and
        -4 for stdev (across cells).

        Note that preview_index is taken from data without checking cell table.
        Caller has to figure out which index along memory cell dimension they
        actually want to preview in case it needs to be a specific cell / pulse.

        Will typically reuse data from corrected output buffer. Therefore,
        correct(...) must have been called with the appropriate flags before
        compute_preview(...).
        """
        if preview_index < -4:
            raise ValueError(f"No statistic with code {preview_index} defined")
        elif preview_index >= self.frames:
            raise ValueError(f"Memory cell index {preview_index} out of range")

        if preview_index >= 0:
            def fun(a):
                return a[preview_index]
        elif preview_index == -1:
            # note: separate from next case because dtype not applicable here
            fun = functools.partial(np.nanmax, axis=0)
        elif preview_index in (-2, -3, -4):
            fun = functools.partial(
                {
                    -2: np.nanmean,
                    -3: np.nansum,
                    -4: np.nanstd,
                }[preview_index],
                axis=0,
                dtype=np.float32,
            )
        # TODO: reuse output buffers
        # TODO: integrate multithreading
        return (fun(in_buffer_view) for in_buffer_view in self.preview_data_views)

    def reshape(self, output_order, out=None):
        """Move axes to desired output order"""
        # TODO: avoid copy
        if output_order == self._corrected_axis_order:
            self.reshaped_data = self.processed_data
        else:
            self.reshaped_data = np.transpose(
                self.processed_data,
                utils.transpose_order(self._corrected_axis_order, output_order),
            )

        if out is None:
            return self.reshaped_data
        else:
            out[:] = self.reshaped_data


class BaseGpuRunner(BaseKernelRunner):
    """Class to handle GPU buffers and execution of CUDA kernels on image data

    All GPU buffers are kept within this class and it is intentionally very stateful.
    This generally means that you will want to load data into it and then do something.
    Typical usage in correct order:

    1. instantiate
    2. load constants
    3. load data and cell table
    4. correct
    5a. reshape (only here does data transfer back to host)
    5b. compute previews (optional)

    repeat from 2. or 3.

    In case no constants are available / correction is not desired, can skip 3 and 4 and
    pass CorrectionFlags.NONE to correct(...). Generally, user must handle which
    correction steps are appropriate given the constants loaded so far.
    """

    # These must be set by subclass
    _kernel_source_filename = None
    _corrected_axis_order = None

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
    ):
        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )
        global cupy
        import cupy
        _src_dir = pathlib.Path(__file__).absolute().parent
        # subclass must define _kernel_source_filename
        with (_src_dir / "kernels" / self._kernel_source_filename).open("r") as fd:
            self._kernel_template = jinja2.Template(fd.read())

        self._init_kernels()

        # reuse buffers for input / output
        self.cell_table_gpu = cupy.empty(self.frames, dtype=np.uint16)
        self.input_data_gpu = cupy.empty(self.input_shape, dtype=input_data_dtype)
        self.processed_data_gpu = cupy.empty(
            self.processed_shape, dtype=output_data_dtype
        )
        self.reshaped_data_gpu = None  # currently not reusing buffer

    def reshape(self, output_order, out=None):
        """Move axes to desired output order and copy to host memory

        The out parameter is passed directly to the get function of GPU array: if
        None, then a new ndarray (in host memory) is returned. If not None, then data
        will be loaded into the provided array, which must match shape / dtype.
        """
        # TODO: avoid copy
        if output_order == self._corrected_axis_order:
            self.reshaped_data_gpu = self.processed_data_gpu
        else:
            self.reshaped_data_gpu = cupy.transpose(
                self.processed_data_gpu,
                utils.transpose_order(self._corrected_axis_order, output_order),
            )

        return self.reshaped_data_gpu.get(out=out)

    def load_data(self, raw_data):
        self.input_data_gpu.set(raw_data)

    def load_cell_table(self, cell_table):
        self.cell_table_gpu.set(cell_table)

    def update_block_size(self, full_block):
        """Set execution grid such that it covers processed_shape with full_blocks

        Execution is scheduled with 3d "blocks" of CUDA threads. Tuning can affect
        performance. Correction kernels are "monolithic" for simplicity (i.e. each
        logical thread handles one entry in output data), so in each dimension we
        parallelize, grid * block >= length to cover all entries.

        Note that individual kernels must themselves check whether they go out of
        bounds; grid dimensions get rounded up in case ndarray size is not multiple of
        block size.

        """
        assert len(full_block) == 3
        self.full_block = tuple(full_block)
        self.full_grid = tuple(
            utils.ceil_div(a_length, block_length)
            for (a_length, block_length) in zip(self.processed_shape, full_block)
        )

    def compute_previews(self, preview_index):
        """See BaseKernelRunner.compute_previews; GPU version uses the same logic (numpy
        calls dispatched via CuPy) and just .get()s results."""

        return (res.get() for res in super().compute_previews(preview_index))
