import enum

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    FLOAT_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    VECTOR_STRING_ELEMENT,
)

from .. import base_calcat, base_correction, base_kernel_runner, schemas, utils
from .._version import version as deviceVersion


class AgipdConstants(enum.Enum):
    ThresholdsDark = enum.auto()
    Offset = enum.auto()
    SlopesPC = enum.auto()
    SlopesFF = enum.auto()
    BadPixelsDark = enum.auto()
    BadPixelsPC = enum.auto()
    BadPixelsFF = enum.auto()


# from pycalibration's enum.py
class AgipdGainMode(enum.IntEnum):
    ADAPTIVE_GAIN = 0
    FIXED_HIGH_GAIN = 1
    FIXED_MEDIUM_GAIN = 2
    FIXED_LOW_GAIN = 3


# note: if this is extended further, bump up the dtype passed to kernel
class CorrectionFlags(enum.IntFlag):
    NONE = 0
    THRESHOLD = 1
    OFFSET = 2
    BLSHIFT = 4
    REL_GAIN_PC = 8
    GAIN_XRAY = 16
    BPMASK = 32
    FORCE_MG_IF_BELOW = 64
    FORCE_HG_IF_BELOW = 128


class AgipdGpuRunner(base_kernel_runner.BaseGpuRunner):
    _kernel_source_filename = "agipd_gpu.cu"
    _corrected_axis_order = "fxy"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
        bad_pixel_mask_value=np.nan,
        gain_mode=AgipdGainMode.ADAPTIVE_GAIN,
        g_gain_value=1,
        mg_hard_threshold=2000,
        hg_hard_threshold=2000,
    ):
        global cupy
        import cupy
        self.gain_mode = gain_mode
        # default gain only matters when not thresholding (missing constant or fixed)
        # note: gain stage (result of thresholding) is 0, 1, or 2
        if self.gain_mode is AgipdGainMode.ADAPTIVE_GAIN:
            self.default_gain = cupy.uint8(gain_mode)
        else:
            self.default_gain = cupy.uint8(gain_mode - 1)
        self.input_shape = (frames, 2, pixels_x, pixels_y)
        self.processed_shape = (frames, pixels_x, pixels_y)
        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )
        self.gain_map_gpu = cupy.empty(self.processed_shape, dtype=np.float32)

        self.map_shape = (self.constant_memory_cells, self.pixels_x, self.pixels_y)
        self.gm_map_shape = self.map_shape + (3,)  # for gain-mapped constants
        self.threshold_map_shape = self.map_shape + (2,)
        # constants
        self.gain_thresholds_gpu = cupy.empty(
            self.threshold_map_shape, dtype=np.float32
        )
        self.offset_map_gpu = cupy.empty(self.gm_map_shape, dtype=np.float32)
        self.rel_gain_pc_map_gpu = cupy.empty(self.gm_map_shape, dtype=np.float32)
        # not gm_map_shape because it only applies to medium gain pixels
        self.md_additional_offset_gpu = cupy.empty(self.map_shape, dtype=np.float32)
        self.rel_gain_xray_map_gpu = cupy.empty(self.map_shape, dtype=np.float32)
        self.bad_pixel_map_gpu = cupy.empty(self.gm_map_shape, dtype=np.uint32)
        self.set_bad_pixel_mask_value(bad_pixel_mask_value)
        self.set_mg_hard_threshold(mg_hard_threshold)
        self.set_hg_hard_threshold(hg_hard_threshold)
        self.set_g_gain_value(g_gain_value)
        self.flush_buffers(set(AgipdConstants))

        self.update_block_size((1, 1, 64))

    @property
    def preview_data_views(self):
        return (
            self.input_data_gpu[:, 0],  # raw
            self.processed_data_gpu,  # corrected
            self.input_data_gpu[:, 1],  # raw gain
            self.gain_map_gpu,  # digitized gain
        )

    def load_thresholds(self, threshold_map):
        # shape: y, x, memory cell, thresholds and gain values
        # note: the gain values are something like means used to derive thresholds
        self.gain_thresholds_gpu.set(
            np.transpose(threshold_map[..., :2], (2, 1, 0, 3)).astype(np.float32)
        )

    def load_offset_map(self, offset_map):
        # shape: y, x, memory cell, gain stage
        self.offset_map_gpu.set(
            np.transpose(offset_map, (2, 1, 0, 3)).astype(np.float32)
        )

    def load_rel_gain_pc_map(self, slopes_pc_map, override_md_additional_offset=None):
        # pc has funny shape (11, 352, 128, 512) from file
        # this is (fi, memory cell, y, x)
        slopes_pc_map = slopes_pc_map.astype(np.float32)
        # the following may contain NaNs, though...
        hg_slope = slopes_pc_map[0]
        hg_intercept = slopes_pc_map[1]
        mg_slope = slopes_pc_map[3]
        mg_intercept = slopes_pc_map[4]
        # TODO: remove sanitization (should happen in constant preparation notebook)
        # from agipdlib.py: replace NaN with median (per memory cell)
        # note: suffixes in agipdlib are "_m" and "_l", should probably be "_I"
        for naughty_array in (hg_slope, hg_intercept, mg_slope, mg_intercept):
            medians = np.nanmedian(naughty_array, axis=(1, 2))
            nan_bool = np.isnan(naughty_array)
            nan_cell, _, _ = np.where(nan_bool)
            naughty_array[nan_bool] = medians[nan_cell]

            too_low_bool = naughty_array < 0.8 * medians[:, np.newaxis, np.newaxis]
            too_low_cell, _, _ = np.where(too_low_bool)
            naughty_array[too_low_bool] = medians[too_low_cell]

            too_high_bool = naughty_array > 1.2 * medians[:, np.newaxis, np.newaxis]
            too_high_cell, _, _ = np.where(too_high_bool)
            naughty_array[too_high_bool] = medians[too_high_cell]

        frac_hg_mg = hg_slope / mg_slope
        rel_gain_map = np.ones(
            (3, self.constant_memory_cells, self.pixels_y, self.pixels_x),
            dtype=np.float32,
        )
        rel_gain_map[1] = rel_gain_map[0] * frac_hg_mg
        rel_gain_map[2] = rel_gain_map[1] * 4.48
        self.rel_gain_pc_map_gpu.set(np.transpose(rel_gain_map, (1, 3, 2, 0)))
        if override_md_additional_offset is None:
            md_additional_offset = (hg_intercept - mg_intercept * frac_hg_mg).astype(
                np.float32
            )
            self.md_additional_offset_gpu.set(
                np.transpose(md_additional_offset, (0, 2, 1))
            )
        else:
            self.override_md_additional_offset(override_md_additional_offset)

    def override_md_additional_offset(self, override_value):
        self.md_additional_offset_gpu.fill(override_value)

    def load_rel_gain_ff_map(self, slopes_ff_map):
        # constant shape: y, x, memory cell
        if slopes_ff_map.shape[2] == 2:
            # TODO: remove support for old format
            # old format, is per pixel only (shape is y, x, 2)
            # note: we should not support this in online
            slopes_ff_map = np.broadcast_to(
                slopes_ff_map[..., 0][..., np.newaxis],
                (self.pixels_y, self.pixels_x, self.constant_memory_cells),
            )
        self.rel_gain_xray_map_gpu.set(np.transpose(slopes_ff_map).astype(np.float32))

    def set_g_gain_value(self, override_value):
        self.g_gain_value = cupy.float32(override_value)

    def load_bad_pixels_map(self, bad_pixels_map, override_flags_to_use=None):
        # will simply OR with already loaded, does not take into account which ones
        if len(bad_pixels_map.shape) == 3:
            if bad_pixels_map.shape == (
                self.pixels_y,
                self.pixels_x,
                self.constant_memory_cells,
            ):
                # BadPixelsFF is not per gain stage - broadcasting along gain dimension
                self.bad_pixel_map_gpu |= cupy.asarray(
                    np.broadcast_to(
                        np.transpose(bad_pixels_map)[..., np.newaxis],
                        self.gm_map_shape,
                    ),
                    dtype=np.uint32,
                )
            elif bad_pixels_map.shape == (
                self.constant_memory_cells,
                self.pixels_y,
                self.pixels_x,
            ):
                # oh, can also be old bad pixels pc?
                self.bad_pixel_map_gpu |= cupy.asarray(
                    np.broadcast_to(
                        np.transpose(bad_pixels_map, (0, 2, 1))[..., np.newaxis],
                        self.gm_map_shape,
                    ),
                    dtype=np.uint32,
                )
            else:
                raise ValueError(
                    f"Unsupported bad pixel map shape: {bad_pixels_map.shape}"
                )
        else:
            self.bad_pixel_map_gpu |= cupy.asarray(
                np.transpose(bad_pixels_map, (2, 1, 0, 3)), dtype=np.uint32
            )

        if override_flags_to_use is not None:
            self.override_bad_pixel_flags_to_use(override_flags_to_use)

    def override_bad_pixel_flags_to_use(self, override_value):
        self.bad_pixel_map_gpu &= cupy.uint32(override_value)

    def set_bad_pixel_mask_value(self, mask_value):
        self.bad_pixel_mask_value = cupy.float32(mask_value)

    def set_mg_hard_threshold(self, value):
        self.mg_hard_threshold = cupy.float32(value)

    def set_hg_hard_threshold(self, value):
        self.hg_hard_threshold = cupy.float32(value)

    def flush_buffers(self, constants):
        if AgipdConstants.Offset in constants:
            self.offset_map_gpu.fill(0)
        if AgipdConstants.SlopesPC in constants:
            self.rel_gain_pc_map_gpu.fill(1)
            self.md_additional_offset_gpu.fill(0)
        if AgipdConstants.SlopesFF:
            self.rel_gain_xray_map_gpu.fill(1)
        if constants & {
            AgipdConstants.BadPixelsDark,
            AgipdConstants.BadPixelsPC,
            AgipdConstants.BadPixelsFF,
        }:
            self.bad_pixel_map_gpu.fill(0)
            self.bad_pixel_map_gpu[
                :, 64:512:64
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
            self.bad_pixel_map_gpu[
                :, 63:511:64
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value

    # TODO: baseline shift

    def correct(self, flags):
        if flags & CorrectionFlags.BLSHIFT:
            raise NotImplementedError("Baseline shift not implemented yet")

        self.correction_kernel(
            self.full_grid,
            self.full_block,
            (
                self.input_data_gpu,
                self.cell_table_gpu,
                cupy.uint8(flags),
                self.default_gain,
                self.gain_thresholds_gpu,
                self.mg_hard_threshold,
                self.hg_hard_threshold,
                self.offset_map_gpu,
                self.rel_gain_pc_map_gpu,
                self.md_additional_offset_gpu,
                self.rel_gain_xray_map_gpu,
                self.g_gain_value,
                self.bad_pixel_map_gpu,
                self.bad_pixel_mask_value,
                self.gain_map_gpu,
                self.processed_data_gpu,
            ),
        )

    def _init_kernels(self):
        kernel_source = self._kernel_template.render(
            {
                "pixels_x": self.pixels_x,
                "pixels_y": self.pixels_y,
                "frames": self.frames,
                "constant_memory_cells": self.constant_memory_cells,
                "input_data_dtype": utils.np_dtype_to_c_type(self.input_data_dtype),
                "output_data_dtype": utils.np_dtype_to_c_type(self.output_data_dtype),
                "corr_enum": utils.enum_to_c_template(CorrectionFlags),
            }
        )
        self.source_module = cupy.RawModule(code=kernel_source)
        self.correction_kernel = self.source_module.get_function("correct")


class AgipdCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = AgipdConstants

    @property
    def _constants_need_conditions(self):
        return {
            AgipdConstants.ThresholdsDark: self.dark_condition,
            AgipdConstants.Offset: self.dark_condition,
            AgipdConstants.SlopesPC: self.dark_condition,
            AgipdConstants.SlopesFF: self.illuminated_condition,
            AgipdConstants.BadPixelsDark: self.dark_condition,
            AgipdConstants.BadPixelsPC: self.dark_condition,
            AgipdConstants.BadPixelsFF: self.illuminated_condition,
        }

    @staticmethod
    def add_schema(schema, managed_keys):
        super(AgipdCalcatFriend, AgipdCalcatFriend).add_schema(
            schema, managed_keys, "AGIPD-Type"
        )

        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(352)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(300)
            .commit()
        )

        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.acquisitionRate")
            .assignmentOptional()
            .defaultValue(1.1)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.gainSetting")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.photonEnergy")
            .assignmentOptional()
            .defaultValue(9.2)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.gainMode")
            .assignmentOptional()
            .defaultValue("ADAPTIVE_GAIN")
            .options(",".join(gain_mode.name for gain_mode in AgipdGainMode))
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.integrationTime")
            .assignmentOptional()
            .defaultValue(12)
            .reconfigurable()
            .commit(),
        )
        managed_keys.add("constantParameters.acquisitionRate")
        managed_keys.add("constantParameters.gainSetting")
        managed_keys.add("constantParameters.photonEnergy")
        managed_keys.add("constantParameters.gainMode")
        managed_keys.add("constantParameters.integrationTime")

        base_calcat.add_status_schema_from_enum(schema, AgipdConstants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Acquisition rate"] = self._get_param("acquisitionRate")

        # TODO: remove this workaround after CalCat update
        integration_time = self._get_param("integrationTime")
        if integration_time != 12:
            res["Integration Time"] = integration_time

        gain_mode = AgipdGainMode[self._get_param("gainMode")]
        if gain_mode is not AgipdGainMode.ADAPTIVE_GAIN:
            res["Gain Mode"] = 1

        # TODO: make configurable whether or not to include gain setting?
        res["Gain Setting"] = self._get_param("gainSetting")

        return res

    def illuminated_condition(self):
        res = self.dark_condition()

        # note: can consider always setting memory cells to 352 for FF
        # (deviation on constants in database should remove need for this, though)

        # for now, FF constants don't care about gain mode
        if "Gain Mode" in res:
            del res["Gain Mode"]

        res["Source Energy"] = self._get_param("photonEnergy")

        return res


@KARABO_CLASSINFO("AgipdCorrection", deviceVersion)
class AgipdCorrection(base_correction.BaseCorrection):
    # subclass *must* set these attributes
    _correction_flag_class = CorrectionFlags
    _correction_steps = (
        # step name (used in schema), flag to enable for kernel, constants required
        ("thresholding", CorrectionFlags.THRESHOLD, {AgipdConstants.ThresholdsDark}),
        ("offset", CorrectionFlags.OFFSET, {AgipdConstants.Offset}),
        (  # TODO: specify that /both/ constants are needed, not just one or the other
            "forceMgIfBelow",
            CorrectionFlags.FORCE_MG_IF_BELOW,
            {AgipdConstants.ThresholdsDark, AgipdConstants.Offset},
        ),
        (
            "forceHgIfBelow",
            CorrectionFlags.FORCE_HG_IF_BELOW,
            {AgipdConstants.ThresholdsDark, AgipdConstants.Offset},
        ),
        ("relGainPc", CorrectionFlags.REL_GAIN_PC, {AgipdConstants.SlopesPC}),
        ("gainXray", CorrectionFlags.GAIN_XRAY, {AgipdConstants.SlopesFF}),
        (
            "badPixels",
            CorrectionFlags.BPMASK,
            {
                AgipdConstants.BadPixelsDark,
                AgipdConstants.BadPixelsPC,
                AgipdConstants.BadPixelsFF,
                None,  # means stay available even without constants loaded
            },
        ),
    )
    _kernel_runner_class = AgipdGpuRunner
    _calcat_friend_class = AgipdCalcatFriend
    _constant_enum_class = AgipdConstants
    _managed_keys = base_correction.BaseCorrection._managed_keys.copy()

    @staticmethod
    def expectedParameters(expected):
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(schemas.xtdf_output_schema())
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.frames")
            .setNewDefaultValue(352)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("cell")
            .commit(),
        )

        (
            OUTPUT_CHANNEL(expected)
            .key("preview.outputRawGain")
            .dataSchema(schemas.preview_schema())
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("preview.outputGainMap")
            .dataSchema(schemas.preview_schema())
            .commit(),
        )

        AgipdCalcatFriend.add_schema(expected, AgipdCorrection._managed_keys)
        # this is not automatically done by superclass for complicated class reasons
        base_correction.add_correction_step_schema(
            expected,
            AgipdCorrection._managed_keys,
            AgipdCorrection._correction_steps,
        )
        base_correction.add_bad_pixel_config_node(
            expected, AgipdCorrection._managed_keys
        )

        # turn off the force MG / HG steps by default
        for step in ("forceMgIfBelow", "forceHgIfBelow"):
            for flag in ("enable", "preview"):
                (
                    OVERWRITE_ELEMENT(expected)
                    .key(f"corrections.{step}.{flag}")
                    .setNewDefaultValue(False)
                    .commit(),
                )
        # additional settings specific to AGIPD correction steps
        (
            FLOAT_ELEMENT(expected)
            .key("corrections.forceMgIfBelow.hardThreshold")
            .description(
                "If enabled, any pixels assigned to low gain stage which would be "
                "below this threshold after having medium gain offset subtracted will "
                "be reassigned to medium gain."
            )
            .assignmentOptional()
            .defaultValue(1000)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("corrections.forceHgIfBelow.hardThreshold")
            .description(
                "Like forceMgIfBelow, but potentially reassigning from medium gain to "
                "high gain based on threshold and pixel value minus low gain offset. "
                "Applied after forceMgIfBelow, pixels can theoretically be reassigned "
                "twice, from LG to MG and to HG."
            )
            .assignmentOptional()
            .defaultValue(1000)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected)
            .key("corrections.relGainPc.overrideMdAdditionalOffset")
            .displayedName("Override md_additional_offset")
            .description(
                "Toggling this on will use the value in the next field globally for "
                "md_additional_offset. Note that the correction map on GPU gets "
                "overwritten as long as this boolean is True, so reload constants "
                "after turning off."
            )
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("corrections.relGainPc.mdAdditionalOffset")
            .displayedName("Value for md_additional_offset (if overriding)")
            .description(
                "Normally, md_additional_offset (part of relative gain correction) is "
                "computed when loading SlopesPC. In case you want to use a different "
                "value (global for all medium gain pixels), you can specify it here "
                "and set corrections.overrideMdAdditionalOffset to True."
            )
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("corrections.gainXray.gGainValue")
            .displayedName("G_gain_value")
            .description(
                "Newer X-ray gain correction constants are absolute. The default "
                "G_gain_value of 1 means that output is expected to be in keV. If "
                "this is not desired, one can here specify the mean X-ray gain value "
                "over all modules to get ADU values out - operator must manually "
                "find this mean value."
            )
            .assignmentOptional()
            .defaultValue(1)
            .reconfigurable()
            .commit(),
        )
        AgipdCorrection._managed_keys |= {
            "corrections.forceMgIfBelow.hardThreshold",
            "corrections.forceHgIfBelow.hardThreshold",
            "corrections.relGainPc.overrideMdAdditionalOffset",
            "corrections.relGainPc.mdAdditionalOffset",
            "corrections.gainXray.gGainValue",
        }

        # mandatory: manager needs this in schema
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("managedKeys")
            .assignmentOptional()
            .defaultValue(list(AgipdCorrection._managed_keys))
            .commit()
        )

    @property
    def input_data_shape(self):
        return (
            self.unsafe_get("dataFormat.frames"),
            2,
            self.unsafe_get("dataFormat.pixelsX"),
            self.unsafe_get("dataFormat.pixelsY"),
        )

    def __init__(self, config):
        super().__init__(config)
        # note: gain mode single sourced from constant retrieval node
        try:
            np.float32(config.get("corrections.badPixels.maskingValue"))
        except ValueError:
            config["corrections.badPixels.maskingValue"] = "nan"

        self._has_updated_bad_pixel_selection = False

    @property
    def bad_pixel_mask_value(self):
        return np.float32(self.unsafe_get("corrections.badPixels.maskingValue"))

    _override_bad_pixel_flags = property(base_correction.get_bad_pixel_field_selection)

    @property
    def _kernel_runner_init_args(self):
        return {
            "gain_mode": self.gain_mode,
            "bad_pixel_mask_value": self.bad_pixel_mask_value,
            "g_gain_value": self.unsafe_get("corrections.gainXray.gGainValue"),
            "mg_hard_threshold": self.unsafe_get(
                "corrections.forceMgIfBelow.hardThreshold"
            ),
            "hg_hard_threshold": self.unsafe_get(
                "corrections.forceHgIfBelow.hardThreshold"
            ),
        }

    @property
    def gain_mode(self):
        return AgipdGainMode[self.unsafe_get("constantParameters.gainMode")]

    @property
    def _override_md_additional_offset(self):
        if self.unsafe_get("corrections.relGainPc.overrideMdAdditionalOffset"):
            return self.unsafe_get("corrections.relGainPc.mdAdditionalOffset")
        else:
            return None

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
    ):
        """Called by input_handler for each data hash. Should correct data, optionally
        compute preview, write data output, and optionally write preview outputs."""
        # original shape: frame, data/raw_gain, x, y

        pulse_table = np.squeeze(data_hash.get("image.pulseId"))
        if self._frame_filter is not None:
            try:
                cell_table = cell_table[self._frame_filter]
                pulse_table = pulse_table[self._frame_filter]
                image_data = image_data[self._frame_filter]
            except IndexError:
                self.log_status_warn(
                    "Failed to apply frame filter, please check that it is valid!"
                )
                return

        try:
            self.kernel_runner.load_data(image_data)
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            self.log_status_warn(
                f"For debugging - cell table: {type(cell_table)}, {cell_table}"
            )
            self.log_status_warn(
                f"For debugging - data: {type(image_data)}, {image_data.shape}"
            )
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.load_cell_table(cell_table)
        self.kernel_runner.correct(self._correction_flag_enabled)
        self.kernel_runner.reshape(
            output_order=self.unsafe_get("dataFormat.outputAxisOrder"),
            out=buffer_array,
        )
        # after reshape, data for dataOutput is now safe in its own buffer
        with self.warning_context(
            "processingState", base_correction.WarningLampType.PREVIEW_SETTINGS
        ) as warn:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ), preview_warning = utils.pick_frame_index(
                self.unsafe_get("preview.selectionMode"),
                self.unsafe_get("preview.index"),
                cell_table,
                pulse_table,
            )
            if preview_warning is not None:
                warn(preview_warning)
        (
            preview_raw,
            preview_corrected,
            preview_raw_gain,
            preview_gain_map,
        ) = self.kernel_runner.compute_previews(preview_slice_index)

        # reusing input data hash for sending
        data_hash.set("image.data", buffer_handle)
        data_hash.set("calngShmemPaths", ["image.data"])

        data_hash.set("image.cellId", cell_table[:, np.newaxis])
        data_hash.set("image.pulseId", pulse_table[:, np.newaxis])

        self._write_output(data_hash, metadata)
        self._write_preview_outputs(
            (
                ("preview.outputRaw", preview_raw),
                ("preview.outputCorrected", preview_corrected),
                ("preview.outputRawGain", preview_raw_gain),
                ("preview.outputGainMap", preview_gain_map),
            ),
            metadata,
        )

    def _load_constant_to_runner(self, constant, constant_data):
        if constant is AgipdConstants.ThresholdsDark:
            if self.gain_mode is not AgipdGainMode.ADAPTIVE_GAIN:
                self.log.INFO("Loaded ThresholdsDark ignored due to fixed gain mode")
                # TODO: set constant status to ignoring
                return
            self.kernel_runner.load_thresholds(constant_data)
        elif constant is AgipdConstants.Offset:
            self.kernel_runner.load_offset_map(constant_data)
        elif constant is AgipdConstants.SlopesPC:
            self.kernel_runner.load_rel_gain_pc_map(constant_data)
            if self._override_md_additional_offset is not None:
                self.kernel_runner.md_additional_offset_gpu.fill(
                    self._override_md_additional_offset
                )
        elif constant is AgipdConstants.SlopesFF:
            self.kernel_runner.load_rel_gain_ff_map(constant_data)
        elif "BadPixels" in constant.name:
            self.kernel_runner.load_bad_pixels_map(
                constant_data, override_flags_to_use=self._override_bad_pixel_flags
            )

    def preReconfigure(self, config):
        super().preReconfigure(config)
        if config.has("corrections.badPixels.maskingValue"):
            # only check if it is valid; postReconfigure will use it
            np.float32(config.get("corrections.badPixels.maskingValue"))

    def postReconfigure(self):
        super().postReconfigure()

        # TODO: move after getting cached update, check if necessary
        if self._override_md_additional_offset is not None:
            self.kernel_runner.override_md_additional_offset(
                self._override_md_additional_offset
            )

        if not hasattr(self, "_prereconfigure_update_hash"):
            return

        update = self._prereconfigure_update_hash

        if update.has("constantParameters.gainMode"):
            self.flush_constants()
            self._update_buffers()

        if update.has("corrections.forceMgIfBelow.hardThreshold"):
            self.kernel_runner.set_mg_hard_threshold(
                self.get("corrections.forceMgIfBelow.hardThreshold")
            )

        if update.has("corrections.forceHgIfBelow.hardThreshold"):
            self.kernel_runner.set_hg_hard_threshold(
                self.get("corrections.forceHgIfBelow.hardThreshold")
            )

        if update.has("corrections.gainXray.gGainValue"):
            self.kernel_runner.set_g_gain_value(
                self.get("corrections.gainXray.gGainValue")
            )

        if update.has("corrections.badPixels.maskingValue"):
            self.kernel_runner.set_bad_pixel_mask_value(self.bad_pixel_mask_value)

        if any(
            path.startswith("corrections.badPixels.subsetToUse")
            for path in update.getPaths()
        ):
            self.log_status_info("Updating bad pixel maps based on subset specified")
            if any(
                update.get(
                    f"corrections.badPixels.subsetToUse.{field.name}", default=False
                )
                for field in utils.BadPixelValues
            ):
                self.log_status_info(
                    "Some fields reenabled, reloading cached bad pixel constants"
                )
                with self.calcat_friend.cached_constants_lock:
                    self.kernel_runner.flush_buffers(
                        {
                            AgipdConstants.BadPixelsDark,
                            AgipdConstants.BadPixelsPC,
                            AgipdConstants.BadPixelsFF,
                        }
                    )
                    for (
                        constant,
                        data,
                    ) in self.calcat_friend.cached_constants.items():
                        if "BadPixels" in constant.name:
                            self._load_constant_to_runner(constant, data)
            self.kernel_runner.override_bad_pixel_flags_to_use(
                self._override_bad_pixel_flags
            )
