import concurrent.futures
import enum
import functools

import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    VECTOR_STRING_ELEMENT,
    Schema,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,
    schemas,
    preview_utils,
    utils,
)
from .._version import version as deviceVersion


_pretend_pulse_table = np.arange(16, dtype=np.uint8)


class JungfrauConstants(enum.Enum):
    Offset10Hz = enum.auto()
    BadPixelsDark10Hz = enum.auto()
    BadPixelsFF10Hz = enum.auto()
    RelativeGain10Hz = enum.auto()


# from pycalibration (TOOD: move to common shared lib)
class JungfrauGainMode(enum.Enum):
    DYNAMIC_GAIN = "dynamicgain"
    DYNAMIC_GAIN_HG0 = "dynamichg0"
    FIX_GAIN_1 = "fixgain1"
    FIX_GAIN_2 = "fixgain2"
    FORCE_SWITCH_HG1 = "forceswitchg1"
    FORCE_SWITCH_HG2 = "forceswitchg2"


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1
    REL_GAIN = 2
    BPMASK = 4
    STRIXEL = 8


class KernelRunnerVersions(enum.Enum):
    GPU = enum.auto()
    CPU = enum.auto()


class JungfrauGpuRunner(base_kernel_runner.BaseGpuRunner):
    _kernel_source_filename = "jungfrau_gpu.cu"
    _corrected_axis_order = "fyx"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
        bad_pixel_mask_value=np.nan,
    ):
        global cupy
        import cupy
        self.input_shape = (frames, pixels_y, pixels_x)
        self.processed_shape = self.input_shape
        self.update_block_size((1, 1, 64))

        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )
        # note: superclass creates cell table with wrong dtype
        self.cell_table_gpu = cupy.empty(self.frames, dtype=np.uint8)
        self.input_gain_stage_gpu = cupy.empty(self.input_shape, dtype=np.uint8)
        self.map_shape = (self.constant_memory_cells, self.pixels_y, self.pixels_x, 3)
        self.offset_map_gpu = cupy.zeros(self.map_shape, dtype=np.float32)
        self.rel_gain_map_gpu = cupy.ones(self.map_shape, dtype=np.float32)
        self.bad_pixel_map_gpu = cupy.zeros(self.map_shape, dtype=np.uint32)
        self.bad_pixel_mask_value = bad_pixel_mask_value

        # strixel support
        self._strixel_out_shape = (frames, 86, 3090)
        self._strixel_block = ((1, 1, 64))
        # note: only executing kernel on lower half of y range, hence 256
        self._strixel_grid = tuple(
            utils.ceil_div(a_length, block_length)
            for (a_length, block_length) in zip(
                (frames, 256, 1024), self._strixel_block
            )
        )
        self._processed_data_regular_gpu = self.processed_data_gpu
        self._processed_data_strixel_gpu = cupy.empty(
            self._strixel_out_shape, dtype=output_data_dtype
        )

    def _init_kernels(self):
        kernel_source = self._kernel_template.render(
            {
                "pixels_x": self.pixels_x,
                "pixels_y": self.pixels_y,
                "frames": self.frames,
                "constant_memory_cells": self.constant_memory_cells,
                "input_data_dtype": utils.np_dtype_to_c_type(self.input_data_dtype),
                "output_data_dtype": utils.np_dtype_to_c_type(self.output_data_dtype),
                "corr_enum": utils.enum_to_c_template(CorrectionFlags),
                "burst_mode": self.burst_mode,
            }
        )
        self.source_module = cupy.RawModule(code=kernel_source)
        self.correction_kernel = self.source_module.get_function("correct")
        self.strixel_transform_kernel = self.source_module.get_function(
            "strixel_transform"
        )

    @property
    def burst_mode(self):
        return self.frames > 1

    @property
    def preview_data_views(self):
        return (self.input_data_gpu, self.processed_data_gpu, self.input_gain_stage_gpu)

    def load_data(self, image_data, input_gain_stage, cell_table):
        """Experiment: loading all three in one function as they are tied"""
        self.input_data_gpu.set(image_data)
        self.input_gain_stage_gpu.set(input_gain_stage)
        if self.burst_mode:
            self.cell_table_gpu.set(cell_table)

    def flush_buffers(self, constants):
        if JungfrauConstants.Offset10Hz in constants:
            self.offset_map_gpu.fill(0)
        if JungfrauConstants.RelativeGain10Hz in constants:
            self.rel_gain_map_gpu.fill(1)
        if constants & {
            JungfrauConstants.BadPixelsDark10Hz,
            JungfrauConstants.BadPixelsFF10Hz,
        }:
            self.bad_pixel_map_gpu.fill(0)
            self.bad_pixel_map_gpu[
                :, :, 255:1023:256
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
            self.bad_pixel_map_gpu[
                :, :, 256:1024:256
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
            self.bad_pixel_map_gpu[
                :, [255, 256]
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value

    def override_bad_pixel_flags_to_use(self, override_value):
        self.bad_pixel_map_gpu &= cupy.uint32(override_value)

    def correct(self, flags):
        self.correction_kernel(
            self.full_grid,
            self.full_block,
            (
                self.input_data_gpu,
                self.input_gain_stage_gpu,
                self.cell_table_gpu,
                cupy.uint8(flags),
                self.offset_map_gpu,
                self.rel_gain_map_gpu,
                self.bad_pixel_map_gpu,
                self.bad_pixel_mask_value,
                self._processed_data_regular_gpu,
            ),
        )
        if flags & CorrectionFlags.STRIXEL:
            self.strixel_transform_kernel(
                self._strixel_grid,
                self._strixel_block,
                (
                    self._processed_data_regular_gpu,
                    self._processed_data_strixel_gpu,
                ),
            )
            self.processed_data_gpu = self._processed_data_strixel_gpu
        else:
            self.processed_data_gpu = self._processed_data_regular_gpu


class JungfrauCpuRunner(base_kernel_runner.BaseKernelRunner):
    _corrected_axis_order = "fyx"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,  # TODO: configurable
        bad_pixel_mask_value=np.nan,
    ):
        self.input_shape = (frames, pixels_y, pixels_x)
        self.preview_shape = (pixels_y, pixels_x)
        self.processed_shape = self.input_shape
        self._strixel_out_shape = (frames, 86, 3090)

        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )

        # not actually allocating, will just point to incoming data
        self.input_data = None
        self.input_gain_stage = None
        self.processed_data = None
        self._processed_data_regular = np.empty(
            self.processed_shape, dtype=output_data_dtype
        )
        self._processed_data_strixel = np.empty(
            self._strixel_out_shape, dtype=output_data_dtype
        )

        # for computing previews faster
        self.thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=3)

        from ..kernels import jungfrau_cython
        self.correction_kernel_single = jungfrau_cython.correct_single
        self.correction_kernel_burst = jungfrau_cython.correct_burst
        self.correction_kernel_strixel = jungfrau_cython.strixel_transform

        self.map_shape = (self.constant_memory_cells, self.pixels_y, self.pixels_x, 3)
        self.offset_map = np.empty(self.map_shape, dtype=np.float32)
        self.rel_gain_map = np.empty(self.map_shape, dtype=np.float32)
        self.bad_pixel_map = np.empty(self.map_shape, dtype=np.uint32)
        self.bad_pixel_mask_value = bad_pixel_mask_value
        self.flush_buffers(set(JungfrauConstants))

    def __del__(self):
        self.thread_pool.shutdown()

    @property
    def preview_data_views(self):
        return (self.input_data, self.processed_data, self.input_gain_stage)

    @property
    def burst_mode(self):
        return self.frames > 1

    def load_data(self, image_data, input_gain_stage, cell_table):
        self.input_data = image_data.astype(np.uint16, copy=False)
        self.input_gain_stage = input_gain_stage.astype(np.uint8, copy=False)
        self.input_cell_table = cell_table.astype(np.uint8, copy=False)

    def flush_buffers(self, constants):
        if JungfrauConstants.Offset10Hz in constants:
            self.offset_map.fill(0)
        if JungfrauConstants.RelativeGain10Hz in constants:
            self.rel_gain_map.fill(1)
        if constants & {
            JungfrauConstants.BadPixelsDark10Hz,
            JungfrauConstants.BadPixelsFF10Hz,
        }:
            self.bad_pixel_map.fill(0)
            self.bad_pixel_map[
                :, :, 255:1023:256
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
            self.bad_pixel_map[
                :, :, 256:1024:256
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
            self.bad_pixel_map[
                :, [255, 256]
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value

    def override_bad_pixel_flags_to_use(self, override_value):
        self.bad_pixel_map &= np.uint32(override_value)

    def correct(self, flags):
        if self.burst_mode:
            self.correction_kernel_burst(
                self.input_data,
                self.input_gain_stage,
                self.input_cell_table,
                flags,
                self.offset_map,
                self.rel_gain_map,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                self._processed_data_regular,
            )
        else:
            self.correction_kernel_single(
                self.input_data,
                self.input_gain_stage,
                flags,
                self.offset_map,
                self.rel_gain_map,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                self._processed_data_regular,
            )

        if flags & CorrectionFlags.STRIXEL:
            self.correction_kernel_strixel(
                self._processed_data_regular, self._processed_data_strixel
            )
            self.processed_data = self._processed_data_strixel
        else:
            self.processed_data = self._processed_data_regular

    def compute_previews(self, preview_index):
        if preview_index < -4:
            raise ValueError(f"No statistic with code {preview_index} defined")
        elif preview_index >= self.frames:
            raise ValueError(f"Memory cell index {preview_index} out of range")

        if preview_index >= 0:
            def fun(a):
                return a[preview_index]
        elif preview_index == -1:
            # note: separate from next case because dtype not applicable here
            fun = functools.partial(np.nanmax, axis=0)
        elif preview_index in (-2, -3, -4):
            fun = functools.partial(
                {
                    -2: np.nanmean,
                    -3: np.nansum,
                    -4: np.nanstd,
                }[preview_index],
                axis=0,
                dtype=np.float32,
            )
        return self.thread_pool.map(fun, self.preview_data_views)


class JungfrauCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = JungfrauConstants

    @property
    def _constants_need_conditions(self):
        return {
            JungfrauConstants.Offset10Hz: self.dark_condition,
            JungfrauConstants.BadPixelsDark10Hz: self.dark_condition,
            JungfrauConstants.BadPixelsFF10Hz: self.dark_condition,
            JungfrauConstants.RelativeGain10Hz: self.dark_condition,
        }

    @staticmethod
    def add_schema(schema, managed_keys):
        super(JungfrauCalcatFriend, JungfrauCalcatFriend).add_schema(
            schema, managed_keys, "jungfrau-Type"
        )

        # set some defaults for common parameters
        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .setNewDefaultValue(1024)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(90)
            .commit(),
        )

        # add extra parameters
        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.integrationTime")
            .displayedName("Integration time")
            .description("Integration time in ms")
            .assignmentOptional()
            .defaultValue(350)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.sensorTemperature")
            .displayedName("Sensor temperature")
            .description("Sensor temperature in K")
            .assignmentOptional()
            .defaultValue(291)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.gainMode")
            .displayedName("Gain mode")
            .description(
                "Detector may be operating in one of several gain modes. In CalCat, "
                "these map to two CalCat parameters: 'Gain Setting' which is 1 for "
                "dynamic gain with HG0 (0 otherwise) and 'Gain mode' which is 1 "
                "for fixed gain (omitted otherwise)."
            )
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.gainSetting")
            .displayedName("Gain setting")
            .description("See description of gainMode")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
        )
        managed_keys.add("constantParameters.integrationTime")
        managed_keys.add("constantParameters.sensorTemperature")
        managed_keys.add("constantParameters.gainMode")
        managed_keys.add("constantParameters.gainSetting")

        base_calcat.add_status_schema_from_enum(
            schema, JungfrauConstants
        )

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Integration Time"] = self._get_param("integrationTime")
        res["Sensor Temperature"] = self._get_param("sensorTemperature")

        if self._get_param("gainMode") != 0:
            # NOTE: always include if CalCat is updated for this
            res["Gain mode"] = self._get_param("gainMode")
        res["Gain Setting"] = self._get_param("gainSetting")

        return res


@KARABO_CLASSINFO("JungfrauCorrection", deviceVersion)
class JungfrauCorrection(base_correction.BaseCorrection):
    _correction_flag_class = CorrectionFlags
    _correction_steps = (
        ("offset", CorrectionFlags.OFFSET, {JungfrauConstants.Offset10Hz}),
        ("relGain", CorrectionFlags.REL_GAIN, {JungfrauConstants.RelativeGain10Hz}),
        (
            "badPixels",
            CorrectionFlags.BPMASK,
            {
                JungfrauConstants.BadPixelsDark10Hz,
                JungfrauConstants.BadPixelsFF10Hz,
                None,
            }
        ),
        (
            "strixel",
            CorrectionFlags.STRIXEL, set()
        )
    )
    _kernel_runner_class = None  # note: set in __init__ based on config
    _calcat_friend_class = JungfrauCalcatFriend
    _constant_enum_class = JungfrauConstants
    _managed_keys = base_correction.BaseCorrection._managed_keys.copy()
    _image_data_path = "data.adc"
    _cell_table_path = "data.memoryCell"

    @staticmethod
    def expectedParameters(expected):
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(schemas.jf_output_schema())
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .setNewDefaultValue(1024)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.frames")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("frame")
            .commit(),

            # JUNGFRAU data is small, can fit plenty of trains in here
            OVERWRITE_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .setNewDefaultValue(2)
            .commit(),
        )
        (
            # support both CPU and GPU kernels
            STRING_ELEMENT(expected)
            .key("kernelType")
            .assignmentOptional()
            .defaultValue(KernelRunnerVersions.GPU.name)
            .options(",".join(kernel_type.name for kernel_type in KernelRunnerVersions))
            .reconfigurable()
            .commit(),
        )
        JungfrauCorrection._managed_keys.add("kernelType")
        base_correction.add_correction_step_schema(
            expected,
            JungfrauCorrection._managed_keys,
            JungfrauCorrection._correction_steps,
        )
        (
            OVERWRITE_ELEMENT(expected)
            .key("corrections.strixel.enable")
            .setNewDefaultValue(False)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("corrections.strixel.preview")
            .setNewDefaultValue(False)
            .commit(),
        )
        JungfrauCalcatFriend.add_schema(expected, JungfrauCorrection._managed_keys)
        base_correction.add_bad_pixel_config_node(
            expected, JungfrauCorrection._managed_keys
        )
        (
            OUTPUT_CHANNEL(expected)
            .key("preview.outputGainMap")
            .dataSchema(schemas.preview_schema())
            .commit(),
        )

        # mandatory: manager needs this in schema
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("managedKeys")
            .assignmentOptional()
            .defaultValue(list(JungfrauCorrection._managed_keys))
            .commit()
        )

    @property
    def input_data_shape(self):
        return (
            self.unsafe_get("dataFormat.frames"),
            self.unsafe_get("dataFormat.pixelsY"),
            self.unsafe_get("dataFormat.pixelsX"),
        )

    @property
    def _warn_memory_cell_range(self):
        # disable warning in normal operation as cell 15 is expected
        return self.unsafe_get("dataFormat.frames") > 1

    @property
    def output_data_shape(self):
        if self._correction_flag_enabled & CorrectionFlags.STRIXEL:
            axis_lengths = {
                "x": 3090,
                "y": 86,
                "f": self.unsafe_get("dataFormat.filteredFrames"),
            }
            return tuple(
                axis_lengths[axis]
                for axis in self.unsafe_get("dataFormat.outputAxisOrder")
            )
        return super().output_data_shape

    @property
    def _kernel_runner_class(self):
        kernel_type = KernelRunnerVersions[self.unsafe_get("kernelType")]
        if kernel_type is KernelRunnerVersions.CPU:
            return JungfrauCpuRunner
        else:
            return JungfrauGpuRunner

    @property
    def _kernel_runner_init_args(self):
        return {
            "bad_pixel_mask_value": self.bad_pixel_mask_value,
        }

    @property
    def bad_pixel_mask_value(self):
        return np.float32(self.unsafe_get("corrections.badPixels.maskingValue"))

    _override_bad_pixel_flags = property(base_correction.get_bad_pixel_field_selection)

    def __init__(self, config):
        super().__init__(config)
        try:
            np.float32(config.get("corrections.badPixels.maskingValue", default="nan"))
        except ValueError:
            config["corrections.badPixels.maskingValue"] = "nan"

        if config.get("runAsStandaloneModule", default=False):
            schema_override = Schema()
            (
                OUTPUT_CHANNEL(schema_override)
                .key("dataOutput")
                .dataSchema(schemas.jf_output_schema(use_shmem_handle=False))
                .commit(),
            )
            preview_utils.PreviewFriend.add_schema(
                schema_override,
                output_channels=["outputRaw", "outputCorrected", "outputGainMap"],
            )
            self.updateSchema(schema_override)

            def aux():
                self._preview_friend = preview_utils.PreviewFriend(
                    self,
                    output_channels=["outputRaw", "outputCorrected", "outputGainMap"],
                )
                self["availableScenes"] = self["availableScenes"] + [
                    "preview:outputRaw",
                    "preview:outputAssembled",
                    "preview:outputGainMap",
                ]

            self.registerInitialFunction(aux)

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
    ):
        if len(cell_table.shape) == 0:
            cell_table = cell_table[np.newaxis]
        try:
            gain_map = data_hash["data.gain"]
            if self.unsafe_get("workarounds.overrideInputAxisOrder"):
                gain_map.shape = self.input_data_shape
            self.kernel_runner.load_data(
                image_data, gain_map, cell_table
            )
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.correct(self._correction_flag_enabled)
        self.kernel_runner.reshape(
            output_order=self.unsafe_get("dataFormat.outputAxisOrder"),
            out=buffer_array,
        )

        with self.warning_context(
            "processingState", base_correction.WarningLampType.PREVIEW_SETTINGS
        ) as warn:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ), preview_warning = utils.pick_frame_index(
                self.unsafe_get("preview.selectionMode"),
                self.unsafe_get("preview.index"),
                cell_table,
                _pretend_pulse_table,
            )
            if preview_warning is not None:
                warn(preview_warning)
        (
            preview_raw,
            preview_corrected,
            preview_gain_map,
        ) = self.kernel_runner.compute_previews(preview_slice_index)

        # reusing input data hash for sending
        if self.unsafe_get("runAsStandaloneModule"):
            # TODO: use shmem for data.gain, too
            data_hash.set(self._image_data_path, buffer_array)
            data_hash.set("calngShmemPaths", [])
        else:
            data_hash.set(self._image_data_path, buffer_handle)
            data_hash.set("calngShmemPaths", [self._image_data_path])

        self._write_output(data_hash, metadata)

        if self.unsafe_get("runAsStandaloneModule"):
            self._preview_friend.maybe_write(
                [preview_raw, preview_corrected, preview_gain_map]
            )
        else:
            self._write_preview_outputs(
                (
                    ("preview.outputRaw", preview_raw),
                    ("preview.outputCorrected", preview_corrected),
                    ("preview.outputGainMap", preview_gain_map),
                ),
                metadata,
            )

    @property
    def _kernel_type(self):
        return KernelRunnerVersions[self.unsafe_get("kernelType")]

    def _load_constant_to_runner(self, constant, constant_data):
        if constant_data.shape[0] == self.get("dataFormat.pixelsX"):
            constant_data = np.transpose(constant_data, (2, 1, 0, 3))
        else:
            constant_data = np.transpose(constant_data, (2, 0, 1, 3))

        if constant is JungfrauConstants.Offset10Hz:
            if self._kernel_type is KernelRunnerVersions.CPU:
                self.kernel_runner.offset_map[:] = constant_data.astype(np.float32)
            else:
                self.kernel_runner.offset_map_gpu.set(constant_data.astype(np.float32))
            if not self.get("corrections.offset.available"):
                self.set("corrections.offset.available", True)
        elif constant is JungfrauConstants.RelativeGain10Hz:
            if self._kernel_type is KernelRunnerVersions.CPU:
                self.kernel_runner.rel_gain_map[:] = constant_data.astype(np.float32)
            else:
                self.kernel_runner.rel_gain_map_gpu.set(
                    constant_data.astype(np.float32)
                )
            if not self.get("corrections.relGain.available"):
                self.set("corrections.relGain.available", True)
        elif constant in (
            JungfrauConstants.BadPixelsDark10Hz,
            JungfrauConstants.BadPixelsFF10Hz,
        ):
            if self._kernel_type is KernelRunnerVersions.CPU:
                self.kernel_runner.bad_pixel_map |= constant_data
            else:
                self.kernel_runner.bad_pixel_map_gpu |= cupy.asarray(constant_data)
            if not self.get("corrections.badPixels.available"):
                self.set("corrections.badPixels.available", True)
            self.kernel_runner.override_bad_pixel_flags_to_use(
                self._override_bad_pixel_flags
            )

        self._update_correction_flags()
        self.log_status_info(f"Done loading {constant.name} to runner")

    def postReconfigure(self):
        super().postReconfigure()

        if not hasattr(self, "_prereconfigure_update_hash"):
            return

        update = self._prereconfigure_update_hash

        if update.has("corrections.strixel.enable"):
            self._lock_and_update(self._update_frame_filter)

        if any(
            path.startswith("corrections.badPixels.subsetToUse")
            for path in update.getPaths()
        ):
            self.log_status_info("Updating bad pixel maps based on subset specified")
            if any(
                update.get(
                    f"corrections.badPixels.subsetToUse.{field.name}", default=False
                )
                for field in utils.BadPixelValues
            ):
                self.log_status_info(
                    "Some fields reenabled, reloading cached bad pixel constants"
                )
                with self.calcat_friend.cached_constants_lock:
                    self.kernel_runner.flush_buffers(
                        {
                            JungfrauConstants.BadPixelsDark10Hz,
                            JungfrauConstants.BadPixelsFF10Hz,
                        }
                    )
                    for (
                        constant,
                        data,
                    ) in self.calcat_friend.cached_constants.items():
                        if "BadPixels" in constant.name:
                            self._load_constant_to_runner(constant, data)
            self.kernel_runner.override_bad_pixel_flags_to_use(
                self._override_bad_pixel_flags
            )

        if self._preview_friend is not None:
            self._preview_friend.reconfigure(update)
