import enum

import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    VECTOR_STRING_ELEMENT,
    Schema,
)

from .. import (
    base_calcat,
    base_kernel_runner,
    schemas,
    preview_utils,
    utils,
)
from .._version import version as deviceVersion
from ..base_correction import (
    BaseCorrection,
    WarningLampType,
    add_correction_step_schema,
)


class LpdConstants(enum.Enum):
    Offset = enum.auto()
    GainAmpMap = enum.auto()
    RelativeGain = enum.auto()
    FFMap = enum.auto()
    BadPixelsDark = enum.auto()
    BadPixelsFF = enum.auto()


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1
    GAIN_AMP = 2
    REL_GAIN = 4
    FF_CORR = 8
    BPMASK = 16


class LpdGpuRunner(base_kernel_runner.BaseGpuRunner):
    _kernel_source_filename = "lpd_gpu.cu"
    _corrected_axis_order = "fxy"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
        bad_pixel_mask_value=np.nan,
    ):
        global cupy
        import cupy
        self.input_shape = (frames, 1, pixels_y, pixels_x)
        self.processed_shape = (frames, pixels_y, pixels_x)
        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )
        self.gain_map_gpu = cupy.empty(self.processed_shape, dtype=np.float32)

        self.map_shape = (constant_memory_cells, pixels_x, pixels_y, 3)
        self.offset_map_gpu = cupy.zeros(self.map_shape, dtype=np.float32)
        self.gain_amp_map_gpu = cupy.ones(self.map_shape, dtype=np.float32)
        self.rel_gain_slopes_map_gpu = cupy.ones(self.map_shape, dtype=np.float32)
        self.flatfield_map_gpu = cupy.ones(self.map_shape, dtype=np.float32)
        self.bad_pixel_map_gpu = cupy.zeros(self.map_shape, dtype=np.uint32)
        self.bad_pixel_mask_value = bad_pixel_mask_value

        self.update_block_size((1, 1, 64))

    @property
    def preview_data_views(self):
        # TODO: always split off gain from raw to avoid messing up preview?
        return (
            self.input_data_gpu[:, 0],  # raw
            self.processed_data_gpu,  # corrected
            self.gain_map_gpu,  # gain (split from raw)
        )

    def correct(self, flags):
        self.correction_kernel(
            self.full_grid,
            self.full_block,
            (
                self.input_data_gpu,
                self.cell_table_gpu,
                cupy.uint8(flags),
                self.offset_map_gpu,
                self.gain_amp_map_gpu,
                self.rel_gain_slopes_map_gpu,
                self.flatfield_map_gpu,
                self.bad_pixel_map_gpu,
                self.bad_pixel_mask_value,
                self.gain_map_gpu,
                self.processed_data_gpu,
            ),
        )

    def load_constant(self, constant_type, constant_data):
        # constant type → transpose order
        bad_pixel_loading = {
            LpdConstants.BadPixelsDark: (2, 1, 0, 3),
            LpdConstants.BadPixelsFF: (2, 0, 1, 3),
        }
        # constant type → transpose order, GPU buffer
        other_constant_loading = {
            LpdConstants.Offset: ((2, 1, 0, 3), self.offset_map_gpu),
            LpdConstants.GainAmpMap: ((2, 0, 1, 3), self.gain_amp_map_gpu),
            LpdConstants.FFMap: ((2, 0, 1, 3), self.flatfield_map_gpu),
            LpdConstants.RelativeGain: ((2, 1, 0, 3), self.rel_gain_slopes_map_gpu),
        }
        if constant_type in bad_pixel_loading:
            self.bad_pixel_map_gpu |= cupy.asarray(
                np.transpose(
                    constant_data,
                    bad_pixel_loading[constant_type],
                ),
                dtype=np.uint32,
            )
        elif constant_type in other_constant_loading:
            transpose_order, gpu_buffer = other_constant_loading[constant_type]
            gpu_buffer.set(
                np.transpose(
                    constant_data.astype(np.float32),
                    transpose_order,
                )
            )
        else:
            raise ValueError(f"Unhandled constant type {constant_type}")

    def _init_kernels(self):
        kernel_source = self._kernel_template.render(
            {
                "pixels_x": self.pixels_x,
                "pixels_y": self.pixels_y,
                "frames": self.frames,
                "constant_memory_cells": self.constant_memory_cells,
                "input_data_dtype": utils.np_dtype_to_c_type(self.input_data_dtype),
                "output_data_dtype": utils.np_dtype_to_c_type(self.output_data_dtype),
                "corr_enum": utils.enum_to_c_template(CorrectionFlags),
            }
        )
        self.source_module = cupy.RawModule(code=kernel_source)
        self.correction_kernel = self.source_module.get_function("correct")

    def flush_buffers(self, constants):
        if LpdConstants.Offset in constants:
            self.offset_map_gpu.fill(0)
        if LpdConstants.GainAmpMap in constants:
            self.gain_amp_map_gpu.fill(1)
        if LpdConstants.RelativeGain in constants:
            self.rel_gain_slopes_map_gpu.fill(1)
        if LpdConstants.FFMap in constants:
            self.flatfield_map_gpu.fill(1)
        if constants & {LpdConstants.BadPixelsDark, LpdConstants.BadPixelsFF}:
            self.bad_pixel_map_gpu.fill(0)


class LpdCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = LpdConstants

    @property
    def _constants_need_conditions(self):
        return {
            LpdConstants.Offset: self.dark_condition,
            LpdConstants.BadPixelsDark: self.dark_condition,
            LpdConstants.GainAmpMap: self.category_condition,
            LpdConstants.FFMap: self.category_condition,
            LpdConstants.RelativeGain: self.category_condition,
            LpdConstants.BadPixelsFF: self.category_condition,
        }

    @staticmethod
    def add_schema(
        schema,
        managed_keys,
    ):
        super(LpdCalcatFriend, LpdCalcatFriend).add_schema(
            schema, managed_keys, "LPD-Type"
        )

        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(250)
            .commit(),
        )

        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.feedbackCapacitor")
            .assignmentOptional()
            .defaultValue(5)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.photonEnergy")
            .assignmentOptional()
            .defaultValue(9.3)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.category")
            .displayedName("Category")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
        )
        managed_keys.add("constantParameters.feedbackCapacitor")
        managed_keys.add("constantParameters.photonEnergy")
        managed_keys.add("constantParameters.category")

        base_calcat.add_status_schema_from_enum(schema, LpdConstants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Feedback capacitor"] = self._get_param("feedbackCapacitor")

        return res

    def illuminated_condition(self):
        res = self.dark_condition()
        res["Source Energy"] = self._get_param("photonEnergy")
        return res

    def category_condition(self):
        res = self.illuminated_condition()
        res["category"] = self._get_param("category")
        return res


@KARABO_CLASSINFO("LpdCorrection", deviceVersion)
class LpdCorrection(BaseCorrection):
    _correction_flag_class = CorrectionFlags
    _correction_steps = (
        ("offset", CorrectionFlags.OFFSET, {LpdConstants.Offset}),
        ("gainAmp", CorrectionFlags.GAIN_AMP, {LpdConstants.GainAmpMap}),
        ("relGain", CorrectionFlags.REL_GAIN, {LpdConstants.RelativeGain}),
        ("flatfield", CorrectionFlags.FF_CORR, {LpdConstants.FFMap}),
        (
            "badPixels",
            CorrectionFlags.BPMASK,
            {
                LpdConstants.BadPixelsDark,
                LpdConstants.BadPixelsFF,
            }
        ),
    )
    _kernel_runner_class = LpdGpuRunner
    _calcat_friend_class = LpdCalcatFriend
    _constant_enum_class = LpdConstants
    _managed_keys = BaseCorrection._managed_keys.copy()

    @staticmethod
    def expectedParameters(expected):
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(schemas.xtdf_output_schema())
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.frames")
            .setNewDefaultValue(512)
            .commit(),

            # TODO: determine ideal default
            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("frame")
            .commit(),
        )

        (
            OUTPUT_CHANNEL(expected)
            .key("preview.outputGainMap")
            .dataSchema(schemas.preview_schema())
            .commit(),
        )

        add_correction_step_schema(
            expected,
            LpdCorrection._managed_keys,
            LpdCorrection._correction_steps,
        )
        LpdCalcatFriend.add_schema(expected, LpdCorrection._managed_keys)

        # additional settings for correction steps
        (
            STRING_ELEMENT(expected)
            .key("corrections.badPixels.maskingValue")
            .displayedName("Bad pixel masking value")
            .description(
                "Any pixels masked by the bad pixel mask will have their value "
                "replaced with this. Note that this parameter is to be interpreted as "
                "a numpy.float32; use 'nan' to get NaN value."
            )
            .assignmentOptional()
            .defaultValue("nan")
            .reconfigurable()
            .commit(),
        )
        LpdCorrection._managed_keys.add("corrections.badPixels.maskingValue")

        # mandatory: manager needs this in schema
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("managedKeys")
            .assignmentOptional()
            .defaultValue(list(LpdCorrection._managed_keys))
            .commit()
        )

    @property
    def input_data_shape(self):
        return (
            self.unsafe_get("dataFormat.frames"),
            1,
            self.unsafe_get("dataFormat.pixelsX"),
            self.unsafe_get("dataFormat.pixelsY"),
        )

    def __init__(self, config):
        super().__init__(config)
        try:
            np.float32(config.get("corrections.badPixels.maskingValue"))
        except ValueError:
            config["corrections.badPixels.maskingValue"] = "nan"

        if config.get("runAsStandaloneModule", default=False):
            schema_override = Schema()
            (
                OUTPUT_CHANNEL(schema_override)
                .key("dataOutput")
                .dataSchema(schemas.xtdf_output_schema(use_shmem_handle=False))
                .commit(),
            )
            preview_utils.PreviewFriend.add_schema(
                schema_override,
                output_channels=["outputRaw", "outputCorrected", "outputGainMap"],
            )
            self.updateSchema(schema_override)

            def aux():
                self._preview_friend = preview_utils.PreviewFriend(
                    self,
                    output_channels=["outputRaw", "outputCorrected", "outputGainMap"],
                )
                self["availableScenes"] = self["availableScenes"] + [
                    "preview:outputRaw",
                    "preview:outputAssembled",
                    "preview:outputGainMap",
                ]

            self.registerInitialFunction(aux)

    def _load_constant_to_runner(self, constant, constant_data):
        self.kernel_runner.load_constant(constant, constant_data)

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
    ):
        pulse_table = np.ravel(data_hash.get("image.pulseId"))
        if self._frame_filter is not None:
            try:
                cell_table = cell_table[self._frame_filter]
                pulse_table = pulse_table[self._frame_filter]
                image_data = image_data[self._frame_filter]
            except IndexError:
                self.log_status_warn(
                    "Failed to apply frame filter, please check that it is valid!"
                )
                return

        try:
            # drop the singleton dimension for kernel runner
            self.kernel_runner.load_data(image_data)
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.load_cell_table(cell_table)
        self.kernel_runner.correct(self._correction_flag_enabled)
        self.kernel_runner.reshape(
            output_order=self.unsafe_get("dataFormat.outputAxisOrder"),
            out=buffer_array,
        )
        with self.warning_context(
            "processingState", WarningLampType.PREVIEW_SETTINGS
        ) as warn:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ), preview_warning = utils.pick_frame_index(
                self.unsafe_get("preview.selectionMode"),
                self.unsafe_get("preview.index"),
                cell_table,
                pulse_table,
            )
            if preview_warning is not None:
                warn(preview_warning)
            (
                preview_raw,
                preview_corrected,
                preview_gain_map,
            ) = self.kernel_runner.compute_previews(preview_slice_index)

        if self.unsafe_get("runAsStandaloneModule"):
            data_hash.set(self._image_data_path, buffer_array)
            data_hash.set(self._cell_table_path, cell_table[:, np.newaxis])
            data_hash.set("image.pulseId", pulse_table[:, np.newaxis])
            data_hash.set("calngShmemPaths", [])
        else:
            data_hash.set(self._image_data_path, buffer_handle)
            data_hash.set(self._cell_table_path, cell_table[:, np.newaxis])
            data_hash.set("image.pulseId", pulse_table[:, np.newaxis])
            data_hash.set("calngShmemPaths", [self._image_data_path])

        self._write_output(data_hash, metadata)

        if self.unsafe_get("runAsStandaloneModule"):
            self._preview_friend.maybe_write(
                [preview_raw, preview_corrected, preview_gain_map]
            )
        else:
            self._write_preview_outputs(
                (
                    ("preview.outputRaw", preview_raw),
                    ("preview.outputCorrected", preview_corrected),
                    ("preview.outputGainMap", preview_gain_map),
                ),
                metadata,
            )

    def preReconfigure(self, config):
        # TODO: DRY (taken from AGIPD device)
        super().preReconfigure(config)
        if config.has("corrections.badPixels.maskingValue"):
            # only check if it is valid (let raise exception)
            # if valid, postReconfigure will use it
            np.float32(config.get("corrections.badPixels.maskingValue"))

    def postReconfigure(self):
        super().postReconfigure()
        if not hasattr(self, "_prereconfigure_update_hash"):
            return

        update = self._prereconfigure_update_hash
        if update.has("corrections.badPixels.maskingValue"):
            self.kernel_runner.bad_pixel_mask_value = self.bad_pixel_mask_value
