import numpy as np
import pytest

from calng.corrections import DsscCorrection

input_dtype = np.uint16
output_dtype = np.float16
corr_dtype = np.float32
pixels_x = 512
pixels_y = 128
memory_cells = 400
offset_map = (
    np.random.random(size=(pixels_x, pixels_y, memory_cells)).astype(corr_dtype) * 20
)
cell_table = np.arange(memory_cells, dtype=np.uint16)
np.random.shuffle(cell_table)
raw_data = np.random.randint(
    low=0, high=2000, size=(memory_cells, pixels_y, pixels_x), dtype=input_dtype
)


# TODO: gather CPU implementations elsewhere
def correct_cpu(data, cell_table, offset_map):
    corr = np.squeeze(data).astype(corr_dtype, copy=True)
    safe_cell_bool = cell_table < offset_map.shape[-1]
    safe_cell_index = cell_table[safe_cell_bool]
    corr[safe_cell_bool] -= offset_map.transpose()[safe_cell_index]
    return corr.astype(output_dtype, copy=False)


corrected_data = correct_cpu(raw_data, cell_table, offset_map)
only_cast_data = np.squeeze(raw_data).astype(output_dtype)


kernel_runner = DsscCorrection.DsscGpuRunner(
    pixels_x,
    pixels_y,
    memory_cells,
    constant_memory_cells=memory_cells,
    input_data_dtype=input_dtype,
    output_data_dtype=output_dtype,
)


def test_only_cast():
    kernel_runner.load_data(raw_data)
    kernel_runner.correct(DsscCorrection.CorrectionFlags.NONE)
    assert np.allclose(
        kernel_runner.processed_data_gpu.get(), raw_data.astype(output_dtype)
    )


def test_correct():
    kernel_runner.load_offset_map(offset_map)
    kernel_runner.load_data(raw_data)
    kernel_runner.load_cell_table(cell_table)
    kernel_runner.correct(DsscCorrection.CorrectionFlags.OFFSET)
    assert np.allclose(kernel_runner.processed_data_gpu.get(), corrected_data)


def test_correct_oob_cells():
    kernel_runner.load_offset_map(offset_map)
    kernel_runner.load_data(raw_data)
    # here, half the cell IDs will be out of bounds
    wild_cell_table = cell_table * 2
    kernel_runner.load_cell_table(wild_cell_table)
    # should not crash
    kernel_runner.correct(DsscCorrection.CorrectionFlags.OFFSET)
    # should correct as much as possible
    assert np.allclose(
        kernel_runner.processed_data_gpu.get(),
        correct_cpu(raw_data, wild_cell_table, offset_map),
    )


def test_reshape():
    kernel_runner.processed_data_gpu.set(corrected_data)
    assert np.allclose(
        kernel_runner.reshape(output_order="xyf"), corrected_data.transpose()
    )


def test_preview_slice():
    kernel_runner.load_data(raw_data)
    kernel_runner.processed_data_gpu.set(corrected_data)
    preview_raw, preview_corrected = kernel_runner.compute_previews(42)
    assert np.allclose(
        preview_raw,
        raw_data[42].astype(np.float32),
    )
    assert np.allclose(
        preview_corrected,
        corrected_data[42].astype(np.float32),
    )


def test_preview_max():
    # note: in case correction failed, still test this separately
    kernel_runner.load_data(raw_data)
    kernel_runner.processed_data_gpu.set(corrected_data)
    preview_raw, preview_corrected = kernel_runner.compute_previews(-1)
    assert np.allclose(preview_raw, np.max(raw_data, axis=0).astype(np.float32))
    assert np.allclose(
        preview_corrected, np.max(corrected_data, axis=0).astype(np.float32)
    )


def test_preview_mean():
    kernel_runner.load_data(raw_data)
    kernel_runner.processed_data_gpu.set(corrected_data)
    preview_raw, preview_corrected = kernel_runner.compute_previews(-2)
    assert np.allclose(preview_raw, np.nanmean(raw_data, axis=0, dtype=np.float32))
    assert np.allclose(
        preview_corrected, np.nanmean(corrected_data, axis=0, dtype=np.float32)
    )


def test_preview_sum():
    kernel_runner.load_data(raw_data)
    kernel_runner.processed_data_gpu.set(corrected_data)
    preview_raw, preview_corrected = kernel_runner.compute_previews(-3)
    assert np.allclose(preview_raw, np.nansum(raw_data, axis=0, dtype=np.float32))
    assert np.allclose(
        preview_corrected, np.nansum(corrected_data, axis=0, dtype=np.float32)
    )


def test_preview_std():
    kernel_runner.load_data(raw_data)
    kernel_runner.processed_data_gpu.set(corrected_data)
    preview_raw, preview_corrected = kernel_runner.compute_previews(-4)
    assert np.allclose(preview_raw, np.nanstd(raw_data, axis=0, dtype=np.float32))
    assert np.allclose(
        preview_corrected, np.nanstd(corrected_data, axis=0, dtype=np.float32)
    )


def test_preview_valid_index():
    with pytest.raises(ValueError):
        kernel_runner.compute_previews(-5)
    with pytest.raises(ValueError):
        kernel_runner.compute_previews(memory_cells)
